h13068
s 00007/00008/00050
d D 1.2 94/06/22 14:59:43 hjs 3 1
c cstyle -p clean
e
s 00000/00000/00000
d R 1.2 94/03/16 14:41:23 Codemgr 2 1
c SunPro Code Manager data about conflicts, renames, etc...
c Name history : 1 0 src/driver_dev/pio/pio_reg.h
e
s 00058/00000/00000
d D 1.1 94/03/16 14:41:22 hjs 1 0
c date and time created 94/03/16 14:41:22 by hjs
e
u
U
f e 0
t
T
I 1
D 3
/* %W% copyright %E% Sun Microsystems, Inc." */
E 3
I 3

#pragma	ident	"%Z%%M% %I%	%E% SMI"

E 3
/*
 * Copyright (c) 1992, Sun Microsystems, Inc.  All Rights Reserved.
 * Sun considers its source code as an unpublished, proprietary
 * trade secret, and it is available only under strict license
 * provisions.  This copyright notice is placed here only to protect
 * Sun in the event the source is deemed a published work.
 * Disassembly, decompilation, or other means of reducing the
 * object code to human readable form is prohibited by the license
 * agreement under which this code is provided to the user or
 * company in possession of this copy.
D 3
 * 
E 3
I 3
 *
E 3
 * RESTRICTED RIGHTS LEGEND: Use, duplication, or disclosure by the
 * Government is subject to restrictions as set forth in
 * subparagraph (c)(1)(ii) of the Rights in Technical Data and
 * Computer Software clause at DFARS 52.227-7013 and in similar
 * clauses in the FAR and NASA FAR Supplement.
D 3
 * 
E 3
I 3
 *
E 3
 * These examples are provided with no warranties of any kind,
 * including without limitation accuracy and usefulness, and Sun
 * expressly disclaims all implied warranties of merchantability,
 * fitness for a particular purpose and non-infringement. In no
 * event shall Sun be liable for any damages, including without
 * limitation, direct, special, indirect, or consequential damages
 * arising out of, or relating to, use of these examples by customer
 * or any third party. Sun is under no obligation to provide support
 * to customer for this software.
 */

#ifdef i386
#define	PIO_NAME	"pio"		/* name of the device we support */
#else
#define	PIO_NAME	"SUNW,pio"	/* name string in FCode prom */
#endif

typedef struct pio {				/* per-unit structure */
	dev_info_t		*dip;
	kmutex_t		mutex;
	kcondvar_t		cv;
	ddi_iblock_cookie_t	iblock_cookie;	/* for mutexes */
	Pio_csr			*csr;		/* mapped to CSR on device */
	u_char			*data;		/* data register on device */
D 3
	u_char			open;		
E 3
I 3
	u_char			open;
E 3
} Pio;

D 3
#define PIO_OPEN	0x01			/* 1 if open; for excl open */



E 3
I 3
#define	PIO_OPEN	0x01			/* 1 if open; for excl open */
E 3

#ifndef	TRUE
#define	TRUE	1
#define	FALSE	0
#endif	TRUE

#ifndef	OK
#define	OK	0
#endif	OK
E 1
